/// \file standalone.c -- VS1053 Stand-Alone MP3 Player Source Code
/// Copyright (C) 2005-2009 VLSI Solution Oy Tampere Finland

// See readme.txt.

/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

- VS1053:
  + I2S
  + higher reference
- IN VS1053b:
  + SCLK = GPIDATA8
  + XCS  = GPIDATA9
  + SI   = GPIDATA10
  + XDCS = GPIDATA11
 */

/*
  Voice-activated recording:
  1) go directly into record mode ('w' is sent),
     but start saving only after threshold
     has been reached and GPIO6 is low.
  2) when saving starts, raise GPIO7 and send 'r'.
  3) if data has been below the threshold for xx seconds, stop recording
  4) If GPIO6 is high, stop recording
  5) when saving stops, drop GPIO7, send 'e' (end), 'c' (created file)
     goto 1)

  When file is created, use the date and time provided through UART.
  (Protocol to be decided.)
 */
 #if 0 /*only voice-activated recorder*/
   #ifdef RECORDER
    #define VOICE_ACTIVATED
    #define NOCHECKFILETYPE
    #define NO_DREQ_LED
     #ifndef SCI_UI
      #define UART_BUFFERED
     #endif /*!SCI_UI*/
   #endif/*RECORDER*/
 #endif /*0*/


//#define NOCHECKFILETYPE 
 
// #define USE_FLAC // PKP: FLAC support needs special library from VLSI PO

#define COMPAT_KEYS /* PKP: Read the old 3-button interface. */

#define SHUFFLE_PLAY /*shuffle instead of random +12/19 words*/
#define SKIP_ID3V2    /* skip ID3v2 tags */

#ifndef NO_WMA
  //#define UI_PAUSE_BEFORE_PLAY //disable pause key function, wait for play
#endif


#define TIMEOUT /* timeout when decoding seems to be stuck */
//#define ENABLE_I2S /*practical for VS1053 only*/
//#define ENABLE_HIGHREF /*VS1053 only -- use if AVDD >= 3.3V */
#define HDATCHECK 800 /* 21words: give up after 800/2=400kB ~ 3-6sec */


#if 0
  #define SAVE_POSITION
  #define SAVE_VOLUME
  #define LOUDNESS_ALWAYS_ON_OFF
#endif

//#define PLAYTIME 15
//#define SAVE_VOLUME /*66 words */
//#define SAVE_POSITION /*32 words*/
#if defined(SAVE_VOLUME) || defined(SAVE_POSITION)
  #define SAVE_STUFF
#endif



#include <string.h>
#include "standalone.h"
#include <stdio.h>

extern volatile s_int16 *srcWp, *srcRp;

extern __y u_int16 g_dcthi[2048];
auto void play_frame(void);
//void Disable();
//void Enable();
//void Reset();

//#undef putchar
//#undef puts

//#include <stdio.h>

#include <stdlib.h>

//#undef ENABLE_WATCHDOG

#if 0 && defined(VS1003) && defined(RECORDER)
  #define LOWER_LOOPBACK_VOL 0x0000
#endif

//#define USE_DEBUG

#if defined(USE_DEBUG) && !defined(UART_BUFFERED)
  #define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC))
  #define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC))
#else
  #define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
  #define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#endif


#if 0
  #define WAIT_FOR_BUTTON /*After poweron or watchdog reset*/
  #define LOUDNESS_ALWAYS_ON_OFF
#endif


#ifndef RECORDER
  #define FIX_WAV_DECODETIME  /* SCI_UI: advance decodetime for short WAV's */
#endif

//#define HDATCHECK 800 /* 21words: give up after 800/2=400kB ~ 3-6sec */


#ifndef NOCHECKFILETYPE
  #define CHECKFILETYPE /* vs1011 29 words? check file type according to short-filename suffix */
#endif/*NOCHECKFILETYPE*/
#define KEEP_DRIVERS_OFF /* .. until MMC located */
#define WITH_SUBDIRECTORIES /* support subdirectories */


void putch(register __a0 int ch);
#ifdef USE_HC
  s_int16 hcShift;
#endif


#ifdef CHECKFILETYPE
  #undef CHECKFILETYPE /*temporary undef*/
  #define CHECKFILETYPE CheckFileType1053
#endif


#ifdef RECORDER /* if standalone player + recorder, disable some features */
  //#undef HDATCHECK           /* no timeout               -14 words */
  //#undef CHECKFILETYPE       /* no filename suffix check -27 words */
  //#undef TIMEOUT
#endif/*RECORDER*/


//#define CTRL3_UPDATE_VOL        (1<<15)
#define CTRL3_BY_NAME           (1<<8) /*locate by name*/
#define CTRL3_RECORD_ON         (1<<7)
#define CTRL3_AT_END            (1<<6)
#define CTRL3_NO_NUMFILES       (1<<5)
#define CTRL3_PAUSE_ON          (1<<4)
#define CTRL3_FILE_READY        (1<<3)

#define CTRL3_PLAY_MODE_MASK    (3<<1)
#define CTRL3_PAUSE_AFTER_PLAY  (3<<1)
#define CTRL3_PAUSE_BEFORE_PLAY (2<<1)
#define CTRL3_LOOP_SONG         (1<<1)
#define CTRL3_PLAY_NEXT         (0<<1)

#define CTRL3_RANDOM_PLAY       (1<<0) /* SCI_AICTRL3 */

#ifndef GPIO_IDATA
  #define GPIO_IDATA GPIO_DATA
#endif

#ifdef SWITCH_CS2_AND_SCLK
  #define SPI_xCS   1 /* GPIO */
  #define SPI_CLK   8
  #define SPI_MISO  4
  #define SPI_MISO_SHIFT -3
  #define SPI_xCS2  2
  #define SPI_MOSI  1 /* DREQ */
#else
  #define SPI_xCS   1 /* GPIO */
  #define SPI_CLK   2
  #define SPI_MISO  4
  #define SPI_xCS2  8
  #define SPI_MOSI  1 /* DREQ */
#endif

//#define SPI_WREN  0x06
//#define SPI_WRDI  0x04
//#define SPI_RDSR  0x05
//#define SPI_WRSR  0x01
#define SPI_READ  0x03
//#define SPI_WRITE 0x02

void puthex(u_int16 d);
#ifdef USE_DEBUG
  static char hex[] = "0123456789ABCDEF";
  void puthex(u_int16 d) {
      register i;
      for (i=0;i<4;i++) {
          putchar(hex[(d>>12)&15]);
          d <<= 4;
      }
  }
  void puthex2(u_int16 d) {
      register i;
      putchar(hex[(d>>4)&15]);
      putchar(hex[(d>>0)&15]);
  }
#endif


typedef enum {
    spi_pause=0,
    spi_file,
    spi_seek,
    spi_waitdata,
    spi_data,
    spi_zeros,
    spi_record,
    spi_waitima
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;
#ifdef TIMEOUT
  u_int16 timeOut;
#endif

struct SPIX { /* do not touch! ASM uses the structure */
    s_int32 size;
    u_int32 sector;
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __y struct FRAGMENT *fragment;
} spix;


#define         MMC_GO_IDLE_STATE   0
#define         MMC_SEND_OP_COND   1
#define         MMC_SEND_IF_COND   8
#define         MMC_SEND_CSD   9
#define         MMC_SEND_CID   10
#define         MMC_SEND_STATUS   13
#define         MMC_SET_BLOCKLEN   16
#define         MMC_READ_SINGLE_BLOCK   17
#define         MMC_WRITE_BLOCK   24
#define         MMC_PROGRAM_CSD   27
#define         MMC_SET_WRITE_PROT   28
#define         MMC_CLR_WRITE_PROT   29
#define         MMC_SEND_WRITE_PROT   30
#define         MMC_TAG_SECTOR_START   32
#define         MMC_TAG_SECTOR_END   33
#define         MMC_UNTAG_SECTOR   34
#define         MMC_TAG_ERASE_GROUP_START   35
#define         MMC_TAG_ERARE_GROUP_END   36
#define         MMC_UNTAG_ERASE_GROUP   37
#define         MMC_ERASE   38
#define         MMC_READ_OCR     58
#define         MMC_CRC_ON_OFF   59
#define         MMC_R1_BUSY   0x80
#define         MMC_R1_PARAMETER   0x40
#define         MMC_R1_ADDRESS   0x20
#define         MMC_R1_ERASE_SEQ   0x10
#define         MMC_R1_COM_CRC   0x08
#define         MMC_R1_ILLEGAL_COM   0x04
#define         MMC_R1_ERASE_RESET   0x02
#define         MMC_R1_IDLE_STATE   0x01
#define         MMC_STARTBLOCK_READ   0xFE
#define         MMC_STARTBLOCK_WRITE   0xFE
#define         MMC_STARTBLOCK_MWRITE   0xFC
#define         MMC_STOPTRAN_WRITE   0xFD
#define         MMC_DE_MASK   0x1F
#define         MMC_DE_ERROR   0x01
#define         MMC_DE_CC_ERROR   0x02
#define         MMC_DE_ECC_FAIL   0x04
#define         MMC_DE_OUT_OF_RANGE   0x04
#define         MMC_DE_CARD_LOCKED   0x04
#define         MMC_DR_MASK   0x1F
#define         MMC_DR_ACCEPT   0x05
#define         MMC_DR_REJECT_CRC   0x0B
#define         MMC_DR_REJECT_WRITE_ERROR   0x0D

#if 1
  auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg);
#else
  auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg) {
      register __b1 u_int16 t;
      #if 1
        puts("Cmd(");
        puthex(cmd);
        puthex(arg>>16);
        puthex(arg);
        puts(")");
      #endif
      /* some MMC brands require the extra clock pulses     */
      SpiSendClocks();
      SpiSendClocks();
      USEX(GPIO_ODATA) &= ~SPI_xCS2;
  
      while (SpiSendReceiveMmc(0xff00,8) != 0xff) {
          /* Unexpected busy signal from MMC */
          #ifdef USE_DEBUG
            putchar('u');
            putchar('\n');
          #endif
      }
      SpiSendReceiveMmc(cmd, 16); /* top half normally 0xff */
      SpiSendReceiveMmc(arg>>16, 16);
      SpiSendReceiveMmc(arg, 16);
      t = SpiSendReceiveMmc(0x9500, 8);    /* Valid CRC for init, then don't care */
  
      cmd = 100;               /* timeout value, try max. 100 times */
      while (cmd && (t & 0x80)) { /* R1 response not detected */
          t = SpiSendReceiveMmc(0xff00, 8);
          cmd--;               /* decrease the timeout value */
      }
      #ifdef USE_DEBUG
        puthex(t);
        putchar('\n');
      #endif
      return t;
  }
#endif

#if 1
  auto void ReadDiskSector(register __a u_int32 sector);
#else
  auto void ReadDiskSector(register __a u_int32 sector) {
      register s_int16 i;
      register u_int16 *buffer = stream_buffer;
  
      #if 0 && defined(USE_DEBUG)
        //puthex(sector>>16);
        puthex(sector);
        puts(" SECTOR");
      #endif

#ifdef USE_HC /* In ASM this uses self-modified code */
      MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector&0x7fffffffUL)<<hcShift);
#else
      MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector<<9));
#endif
    do {
        i = SpiSendReceiveMmc(0xff00,8);
        #if 0 && defined(USE_DEBUG)
          puthex(i);
          puts(" wait");
        #endif
    } while (i == 0xff);

    #if 0
      if (i != 0xfe) {
          SpiSendClocks();/*USEX(GPIO_ODATA) |= SPI_xCS2;*/
          #ifdef USE_DEBUG
            puts("wfd fail");
          #endif
          return;
      }
    #endif
    
    for (i=0; i<512; i+=2) {
        *buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    SpiSendClocks();
    return;
}
#endif

#if 0
  struct Fat {
      byte BS_jmpBoot[3];         /**< 0 x86 Boot Jump Code */
      byte BS_OEMName[8];         /**< 3 Formatter's name, usually "MSWIN4.1" */
      
      word BPB_BytsPerSec;        /**< 11 Bytes per sector (512) */
      byte BPB_SecPerClus;        /**< 13 Sectors per Cluster (1,2,4,8,..,128) */
      word BPB_RsvdSecCnt;        /**< 14 Reserved sectors (1 (32 for FAT32)) */
      byte BPB_NumFATs;           /**< 16 Number of FATs (2) */
      word BPB_RootEntCnt;        /**< 17 FAT12/16 n of root dir entries */
      word BPB_TotSec16;          /**< 19 Old sector count (0 for FAT32) */
      byte BPB_Media;             /**< 21 Media Type (eg 0xF8) */
      word BPB_FATSz16;           /**< 22 Size of one FAT16 in sectors */
      word BPB_SecPerTrk;         /**< 24 Old CHS Sectors Per Track */
      word BPB_NumHeads;          /**< 26 Old CSH Number of Heads */
      u_32 BPB_HiddSec;           /**< 28 n of sectors before this volume */
      u_32 BPB_TotSec32;          /**< 32 New sector count (0 for FAT12/16) */
     
      /** FAT type specific extensions */
      union Extensions{   
  
        /** FAT12/16 specific extensions to Bios Parameter Block*/
        struct Fat16Specific {    
          byte BS_DrvNum;         /**< 36 DOS INT13 Drive Number (0x80=HD)*/
          byte BS_Reserved1;      /**< 37 For WINNT; Format to 0 */
          byte BS_BootSig;        /**< 38 0x29 if next 3 fields are present */
          byte BS_VolID[4];       /**< 39 Volume ID (usually format datetime) */
          byte BS_VolLab[11];     /**< 43 Volume Label */
          byte BS_FilSysType[8];  /**< 54 Decorative name of fs, eg "FAT16   "*/
        } _16;
  
        /** FAT32 specific extensions to Bios Parameter Block*/
        struct Fat32Specific {    
          u_32 BPB_FATSz32;       /**< 36 Size of one FAT32 in sectors */
          word BPB_ExtFlags;      /**< 40 Flags; active FAT number etc */
          word BPB_FSVer;         /**< 42 File System Version (0x0000) */
          u_32 BPB_RootClus;      /**< 44 Start cluster of Root Dir. (2) */
          word BPB_FSInfo;        /**< 48 Start sector of FSINFO in Resvd area */
          word BPB_BkBootSec;     /**< 50 Sector in Resvd area for BkBoot (6) */
          byte BPB_Reserved[12];  /**< 52 Reserved, Always 0. */
          byte BS_DrvNum;         /**< 64 DOS INT13 Drive Number (0x80=HD) */
          byte BS_Reserved1;      /**< 65 For WINNT, Format to 0 */
          byte BS_BootSig;        /**< 66 0x29 if next 3 fields are present */
          byte BS_VolID[4];       /**< 67 Volume ID (usually format datetime) */
          byte BS_VolLab[11];     /**< 71 Volume Label */
          byte BS_FilSysType[8];  /**< 82 Decorative name of fs, eg "FAT32   "*/
        } _32;
      } ext;
    } fat;
#endif

#if 1
  auto u_int16 GetByte(register __c0 u_int16 n);
#else
  auto u_int16 GetByte(register __c0 u_int16 n) {
      register u_int16 *p = stream_buffer[n>>1];
      if (n & 1) {
          n = *p;
      } else {
          n = *p >> 8;
      }
      return n & 0xff;
  }
#endif
#if 1
  auto u_int16 GetWord(register __c0 u_int16 n);
#else
  auto u_int16 GetWord(register __c0 u_int16 n) {
      return GetByte(n) | (GetByte(n+1)<<8);
  }
#endif
#if 1
  auto u_int32 GetLong(register __c0 u_int16 n);
#else
  auto u_int32 GetLong(register __c0 u_int16 n) {
      return GetWord(n) | ((u_int32)GetWord(n+2)<<16);
  }
#endif

extern __y u_int16 fileName[6];
#ifdef FATINFO_IN_Y
  __y struct FATINFO fatInfo;
#else
  struct FATINFO fatInfo;
#endif

#if 0 //def USE_DEBUG
  void PrintDiskSector(void) {
      register int i;
      register int j;
      for (i=0;i<512;i+=16) {
          puthex(i);
          putchar(' ');
          putchar(' ');
          for (j=0;j<16;j++) {
              puthex2(GetByte(i+j));
              putchar(' ');
          }
          putchar(' ');
          for (j=0;j<16;j++) {
              register int t = GetByte(i+j);
              if ((t & 0x7f) < 0x20)
                  putchar('?');
              else
                  putchar(t);
          }
          putchar('\n');
      }
  }
#endif

#if 1 //&& !defined(USE_DEBUG)
  auto u_int16 InitFileSystem(void);
#else
  auto u_int16 InitFileSystem(void) {
      /* Load MBR -- the first sector on disk */
      ReadDiskSector(0);
  
      #if 0 && defined(USE_DEBUG)
        /* Ok, it should be a MBR sector. Let's verify */
        if (stream_buffer[255] != 0x55aa) {
        #ifdef USE_DEBUG
      //      puthex(stream_buffer[255]);
              puts("!MBR");
        #endif
            return -1; /* sector 0 is not MBR. */
        }
      #endif

#if 0 && defined(USE_DEBUG)
      /* This checks that partition 1 is active. Alter code to allow
         other partition configurations. */
      if (GetByte(0x1be) == 0x80) {
      } else {
          puthex(GetByte(0x1be));
          puts(" PF!");
      }
#endif

    /* Now leave MBR and load sector 0 of partition */
    /* check for boot sector instead of partition table */
    if (GetLong(0x36)!= (((long)'1'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F') &&
        GetLong(0x52)!= (((long)'3'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F')
        ) {
        #ifdef USE_DEBUG
          PrintDiskSector();
          puts("");
        #endif
        fatInfo.currentSector = GetLong(0x1c6);
        ReadDiskSector(fatInfo.currentSector);
    }
    #ifdef USE_DEBUG
      PrintDiskSector();
    #endif
    
    {
        u_int16 BPB_BlksPerSec;

        /* Determine FAT Type (16/32) */
        /* This should be done better, but it'll do for now. */
        fatInfo.BPB_RootEntCnt = GetWord(17);
        /* FAT32 does not have separate root entries. */
        /* TODO: should detect FAT32 by BPB_FATSz16==0 */
        fatInfo.IS_FAT_32 = fatInfo.BPB_RootEntCnt ? 0 : 1;
        fatInfo.FilSysType = GetWord(54+3); /* 0x3231 for FAT12 */
        #ifdef USE_DEBUG
          puthex(fatInfo.FilSysType);
          puts("fsys");
        #endif
        #if 0
          if (!IS_FAT_32){
              if (GetByte(54+4)=='2') {/*diskSect.fat.ext._16.BS_FilSysType[4]=='2'*/) {
                  fatInfo.IS_FAT_12 = GetByte(54);
              }
          }
        #endif
        
        #if 0 && defined(USE_DEBUG)
          if (fatInfo.IS_FAT_32) {
              puts("FAT32");
          } else {
              puthex(fatInfo.BPB_RootEntCnt);
              puts(" FAT");
              puthex(GetWord(19));
              puts(" TSc16");
          }
          puthex(GetWord(54+3)); /* 3231 == '12' 3631= '16' 0000=FAT32 */
          puts(" fst");
        #endif
        #if 1 && defined(USE_DEBUG)
          putchar('\"');
          {
              register int i;
              for (i=0; i<11; i++) {
                  if (fatInfo.IS_FAT_32) {
                      putchar(GetByte(71+i));
                  } else {
                      putchar(GetByte(43+i));
                  }
              }
          }
          putchar('\"');
          putchar('\n');
        #endif
        /* OK, let's calculate */
        /* First, let's get rid of the idea that we have byte addresses
           in the file system. Nope, let's only deal in physical disk
           sectors of 512 bytes. First we convert the FAT byter per sector
           value to "512B disk sectors per fat sector" value. */

        BPB_BlksPerSec = GetWord(11) >> 9;

        /* Then we adjust the Sector per Cluster to mean physical disk
           sectors. in 99% of the cases it is already so because bytes
           per sector almost always is 512 in FAT. Maximum cluster size
           is 65536 bytes (128 disk sectors). */

        fatInfo.fatSectorsPerCluster = GetByte(13)/*BPB_SecPerClus*/ * BPB_BlksPerSec;
        #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.fatSectorsPerCluster);
          puts(" FatSPC");
        #endif
        fatInfo.fatStart = fatInfo.currentSector +
            (u_int32)GetWord(14)/*BPB_RsvdSecCnt*/ * BPB_BlksPerSec;
            #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.fatStart);
          puts(" fatS");
            #endif
        fatInfo.rootStart = GetWord(22);/*BPB_FATSz16*/
        if (fatInfo.rootStart == 0) {
            if (!fatInfo.IS_FAT_32) 
                return 0x0b; /* should be FAT32; can not find root directory */
            fatInfo.rootStart = GetLong(36);/*BPB_FATSz32*/
        }
        /* BPB_NumFATs * BPB_BlksPerSec fits in a word (8+7 bits) */
        fatInfo.rootStart = fatInfo.rootStart * (GetByte(16)/*BPB_NumFATs*/ *
                                                 BPB_BlksPerSec)
            + fatInfo.fatStart;
            #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.rootStart);
          puts(" rootS");
            #endif
        fatInfo.dataStart = fatInfo.rootStart + (fatInfo.BPB_RootEntCnt >> 4)
            - (fatInfo.fatSectorsPerCluster << 1); /*first cluster is cluster 2*/

            #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.dataStart);
          puts(" dataS");
  //      puthex(fatInfo.clusters);
  //      puts(" Clustrs");
            #endif
    }
    return 0;
}
#endif

#if 0
  /** FAT/VFAT directory record union */
  union DirRecordUnion {
    /** Standard directory entry */
    struct Entry {
      byte Name[11]; 0
      byte Attr;     11
      byte NTRes;    12
      byte CrtTimeTenth; 13
      word CrtTime;  14
      word CrtDate;  16
      word LstAccDate; 18
      word FstClusHi; 20
      word WrtTime;   22
      word WrtDate;   24
      word FstClusLo; 26
      u_32 FileSize;  28
    } entry;
  
    /** Extended directory entry */
    struct LongEntry {
      byte Ord;      0
      word Name1[5]; 1 /**< characters 1-5 */
      byte Attr;     11
      byte Type;     12 /**< entry type, zero=long name component */
      byte Chksum;   13
      word Name2[6]; 14 /**< characters 6-11 */
      word FstClusLO; 26 /**< zero for long entry */
      word Name3[2]; 28 /**< characters 12-13 */
    } longentry;
  };
#endif


//#define MAX_FRAGMENTS 28//16//22
__y struct FRAGMENT fragments[MAX_FRAGMENTS];

#if 1
  auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
                            register __b u_int32 fatCluster);
#else
  auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
                            register __b u_int32 fatCluster) {
      register __d0 u_int16 offset;
  
      /* fatCluster means now current cluster, starting from 0, not 2. */
      /* note that startcluster=2 is already compensated in fatInitGlobals */
      #if 0 && defined(USE_DEBUG)
        puthex(fatCluster>>16);
        puthex(fatCluster);
        puts(" fatC");
      #endif
      #if 0
        frag->size = 0;
        frag->start = fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;
      #endif
      goto l55; /* read first sector */
  
      while (1) {
          u_int32 t;
          if (fatInfo.IS_FAT_32) {
              t = GetLong(offset) & 0x0fffffffUL;
          } else {
          #if 0 && defined(USE_DEBUG)
                {
                    register int i;
                    for (i=0;i<256;i++) {
                        t = GetByte(i);
                        puthex(t);
                        putchar(' ');
                        if ((i&7)==7)
                            putchar('\n');
                    }
                }
          #endif

            t = GetWord(offset);
            if (t == 0xffffU)
                t = 0x0fffffffUL;
        }
        #if 0 && defined(USE_DEBUG)
          puthex(t>>16);
          puthex(t);
          puts(" c");
        #endif
        frag->size += fatInfo.fatSectorsPerCluster;
        ++fatCluster;
        if (frag->size >= 32768U || t != fatCluster) {
            /* if not sequential cluster, start next fragment */
            #if 0 && defined(USE_DEBUG)
              puts(" frag");
            #endif
            fatCluster = t;
            if (frag == &fragments[MAX_FRAGMENTS-1] ||
                t == 0x0fffffffUL) {
                frag->start |= LAST_FRAGMENT;
                return frag+1;
            }
            frag++;
        l55:
            frag->size = 0;
            frag->start =
                fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;

        } else {
            offset += (fatInfo.IS_FAT_32 ? 4 : 2);
            offset &= 511;
            if (offset != 0)
                continue;
        }
read:
        /* If FAT page is not already in memory, load it. */
        if (fatInfo.IS_FAT_32) {
            offset = ((u_int16)fatCluster & 0x7F) << 2;
            ReadDiskSector((fatCluster >> 7) + fatInfo.fatStart);
        } else {
            offset = ((u_int16)fatCluster & 0xff) << 1;
            ReadDiskSector((fatCluster >> 8) + fatInfo.fatStart);
        }
    }
}
#endif

u_int16 gFileNum[2];
#ifdef RECORDER
  struct REC {
      u_int16 record;
      u_int16 rd;
      u_int16 wr;
      s_int16 fill;
  } record;
  #ifdef VOICE_ACTIVATED
    u_int16 recordThreshold =  4096; /* only 2^n */
    u_int16 recordTimeout   = 16*10; /* 10 seconds */
    u_int16 recordCount     = 0;
    u_int16 recordStarted   = 0;
  #endif
#endif/*RECORDER*/
extern __y s_int16 mallocAreaY[9216];

#ifndef VOICE_ACTIVATED
  #ifdef __VSDSP__
    auto s_int16 OpenFile(register __c0 u_int16 fileNum);
  #else
    auto s_int16 OpenFile(register __c0 u_int16 fileNum) {
        register __i1 __y struct FRAGMENT *curFragment = &fragments[0];
        register __i0 __y struct FRAGMENT * __y nextFragment;
    
        /* Start at the start of root directory. */
    
        if (fatInfo.IS_FAT_32) {
            nextFragment = FragmentList(curFragment, 2);
        } else {
            curFragment->start = fatInfo.rootStart | LAST_FRAGMENT;
            curFragment->size  = (fatInfo.BPB_RootEntCnt >> 4); /* /sectorsPerCluster?*/
            nextFragment = curFragment + 1;
        }
        gFileNum[0] = fileNum;
        gFileNum[1] = 0;
        return HandleDir(curFragment, nextFragment);
    }
  #endif
#endif /*!VOICE_ACTIVATED*/

#ifdef CHECKFILETYPE
  #if 0
    #define MKID(a,b,c) ((a)|((b)<<8)|((u_int32)(c)<<16))
    auto s_int16 CheckFileType(register __a u_int32 name) {
        name &= 0x00ffffffUL;
        if (name == MKID('W','A','V'))
            return -1;
        if (name == MKID('M','P','3'))
            return -1;
        if (name == MKID('M','I','D'))
            return -1;
        if (name == MKID('W','M','A'))
            return -1;
        if (name == MKID('A','S','F'))
            return -1;
        if (name == MKID('A','A','C'))
            return -1;
        if (name == MKID('M','P','4'))
            return -1;
        if (name == MKID('M','4','A'))
            return -1;
        if (name == MKID('3','G','P'))
            return -1;
        if (name == MKID('3','G','2'))
            return -1;
        if (name == MKID('O','G','G'))
            return -1;
        return 0;
    }
  #endif/*1*/
#endif/*CHECKFILETYPE*/

#ifndef VOICE_ACTIVATED
  /*147 words*/
  auto s_int16 HandleDir(register __i2 __y struct FRAGMENT *curFragment,
                         __y struct FRAGMENT *nextFragment) {
      register __c1 int i = 0;
      register u_int32 currentSector = curFragment->start /*& 0x7fffffffUL */;
  
      #ifdef ENABLE_WATCHDOG
        USEX(WDOG_RESET) = WDOG_RESET_VAL;
      #endif
      goto first;
  
      while (1) {
          register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */
  
          /* We are now looking at FAT directory structure. */
          /* Is current file a regular file? */
          if (fn == 0)
              goto eod; /*end of directory*/
          if (fn != 0xe5) { /* Not deleted */
          #ifdef WITH_SUBDIRECTORIES
                /* Is it a subdirectory? */
                if ((GetByte(i+11/*Attr*/) & 0x10) && fn != '.') {
                    /* TODO: do not call if FAT12! */
                    if (HandleDir(nextFragment,
                                  FragmentList(nextFragment,
                                               ((u_int32)GetWord(i+20)<<16)+GetWord(i+26))) < 0)
                        return -1;
                    ReadDiskSector(currentSector);/*reread sector*/
                } else /* it was a subdirectory */
          #endif
              /* Attributes: NO directory, NO volume id, NO system, NO hidden */
              if ((GetByte(i+11/*Attr*/) & 0xde) == 0
              #ifdef CHECKFILETYPE
                    /*
                      check supported files from name suffix:
                      vs1011/vs1002:  wav mp3
                      vs1003:         wav mp3 mid wma asf
                      vs1033: wav mp3 mid wma asf mp4 aac m4a
                    */
                    && CHECKFILETYPE(GetLong(i+8))
              #endif
                  ) {
  
                  /* It is a regular file. */
  
                  #if defined(SCI_UI)
                    /* If open-by-name, check name and open file if match. */
                    if (USEX(SCI_AICTRL3) & CTRL3_BY_NAME) {
                        register int j, t = 0;
                        register u_int16 *p = stream_buffer + i/2;
                        register __y u_int16 *d = fileName;
                        p[5] &= 0xff00; /* clear the attribute byte */
                        for (j=0; j<6; j++) {
                            t |= *d++ ^ *p++;
                        }
                        if (t == 0) {
                            USEX(SCI_AICTRL0) = gFileNum[1];
                            goto thisfile;
                        }
                    }
                  #endif

                gFileNum[1]++; /* total file count */
                if (gFileNum[0]-- == 0) {
                thisfile:
                #ifdef SCI_UI
                      /* Put the 8.3-character filename to memory so the user
                         can read it. Two bytes per word, big-endian. */
                      {   /* 13 words */
                          register int j;
                          register u_int16 *p = stream_buffer + i/2;
                          register __y u_int16 *d = fileName;
                          for (j=0;j<6;j++) {
                              *d++ = *p++;
                          }
                      }
                #endif
                    /* ------------ FILE FOUND ------------- */
                    fatInfo.fileSize = GetLong(i+28); /*do first!*/
                    FragmentList(&fragments[0],
                                 ((u_int32)GetWord(i+20)<<16) + GetWord(i+26));
                    return -1; /* File found, All OK return */
                }
            } /* normal file */
        } /* 0xe5 / 0x00 deleted or unused entry */
        i = (i + 32) & 511;
        if (i == 0) {
            currentSector++;
            if (--(curFragment->size) == 0) {
                if ((s_int16)(curFragment->start>>16) < 0) {
        eod:
        #if defined(SCI_UI)
                      USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME;
        #endif
                    return gFileNum[1];
                }
                curFragment++;
                currentSector = curFragment->start /*& 0x7fffffffUL*/;
            }
        first:
        #if 0 && defined(USE_DEBUG)
              puthex(currentSector);
              puts(" read");
        #endif
            ReadDiskSector(currentSector);
        }
    }
}
#endif/*!VOICE_ACTIVATED*/


#ifdef HDATCHECK
  u_int16 hdatCheck;
#endif
extern __y u_int16 hwSampleRate;

/* Can be done for all formats, because IdleHook is run as the main thread
   and not in interrupts. Whenever IdleHook is called, there either is
   not enough stream data anyway, or the decoder is outputting samples
   and there is not enough space in the audio buffer (thus not reading
   stream). */
u_int16 __x * __x stream_save;

#ifndef SCI_UI
  __y struct {
      u_int16 Data; /* must be first entry! Used by c.s */
      u_int16 Cnt;
      u_int16 Strobe;
      u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
      u_int16 Key;
      s_int16 KeyTime;
  } ui;
#endif

#if 1
  
    #ifdef RECORDER
      #ifdef ENABLE_UI
        s_int16 ApplAddrRec(register __i0 s_int16 **d, register __a1 s_int16 mode,
                            register __a0 s_int16 n);
      #endif/*ENABLE_UI*/
      #ifdef SCI_UI
        s_int16 ApplAddrSciRec(register __i0 s_int16 **d, register __a1 s_int16 mode,
                               register __a0 s_int16 n);
      #endif/*SCI_UI*/
    #endif/*RECORDER*/
  
  #ifdef ENABLE_UI
    s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
                     register __a0 s_int16 n);
    s_int16 ApplAddrButRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
                     register __a0 s_int16 n);
  #endif/*ENABLE_UI*/
  #ifdef SCI_UI
    s_int16 ApplAddrSci(register __i0 s_int16 **d, register __a1 s_int16 mode,
                        register __a0 s_int16 n);
    s_int16 ApplAddrRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
                        register __a0 s_int16 n);
  #endif/*SCI_UI*/
#else/*1*/
  #ifdef ENABLE_UI
    s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
                     register __a0 s_int16 n) {
        /*Don't check mode, all we get is APPL_AUDIO.*/
        //if (mode == APPL_AUDIO)
        {
            register u_int16 t = hwSampleRate/16;
            /* generate a timebase for User Interface */
            ui.Cnt += n;
            if (ui.Cnt > t) {
                ui.Cnt -= t;
                ui.Timer++;
                ui.Strobe = 1;
            }
        }
    }
  #endif/*ENABLE_UI*/
  #ifdef SCI_UI
    s_int16 ApplAddrSci(register __i0 s_int16 **d, register __a1 s_int16 mode,
                     register __a0 s_int16 n) {
        /* Read memory through SCI */
        if (mode == APPL_R2 /*read AICTRL2*/) {
            register s_int16 a = siAddr++;
            if (a - 0x4000 < 0)
                USEX(SCI_AICTRL2) = USEX(a);
            else
                USEX(SCI_AICTRL2) = USEY(a - 0x4000);
        }
        return n;
    }
  #endif/*SCI_UI*/
#endif


#if 1
  auto u_int32 NextSector(void);
#else
  auto u_int32 NextSector(void) {
      spix.sector++;
      if (spix.fragSize-1 == 0) {
          if ((s_int32)spix.sector < 0) /* was last fragment */
              return -1;
          spix.fragment++;
          spix.sector  = spix.fragment->start /*& 0x7fffffffUL*/;
          spix.fragSize = spix.fragment->size; /* *fatInfo.sectorsPerCluster*/
      } else {
          spix.fragSize = spix.fragSize-1;
      }
      return 0;
  }
#endif

#ifdef RECORDER
  #if 0
    /* Note: sector is not logical sector number, it is directly MMC argument */
    auto void WritePhysicalSectorY(register __d u_int32 sector,
                                   register __i3 __y u_int16 *dataBufPtr) {
        register u_int16 c;
    
        c = MmcCommand(MMC_WRITE_BLOCK|0x40, sector & 0x7fffffff);
        while (c != 0x00) {
            //wait for BUSY token, if you get 0x01(idle), it's an ERROR!
            c = SpiSendReceiveMmc(0xff00, 8);
        }
        //MMC_XCS = MMC_SELECTED;
        SpiSendReceiveMmc(0xfffe, 16);
        for (c=0;c<256;c++) {
            SpiSendReceiveMmc(*dataBufPtr++, 16);
        }
        SpiSendReceiveMmc(0xffff, 16); /* send CRC */
    
        //c = SPIGetChar(); //This prints xxx00101, (usually e5) when data ok
        #ifdef USE_DEBUG
          c = SpiSendReceiveMmc(0xff00, 8);
          puthex2(c);
          puts(" response");
        #endif
        #if 0
          while ( SpiSendReceiveMmc(0xff00, 8) != 0xff)
            ;
          SpiSendReceiveMmc(0xffff, 16);
          SpiSendReceiveMmc(0xffff, 16);
          SpiSendClocks();
          SpiSendClocks();
        #endif
    } 
  #endif
#endif/*RECORDER*/

#if 1
  auto u_int16 swap(register __a0 u_int16 val);
#else
  auto u_int16 swap(register __a0 u_int16 val) {
      register u_int32 v = val<<8;
      return (u_int16)(v>>16)|(u_int16)v;
  }
#endif


#if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
  u_int16 lastDecodeTime;
#endif /*FIX_WAV_DECODETIME + SCI_UI*/

void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#ifdef SAVE_VOLUME
  int volChanged = 0;
#endif

#ifdef SAVE_STUFF
  #define SPI_WREN  0x06
  #define SPI_WRDI  0x04
  #define SPI_RDSR  0x05
  #define SPI_WRSR  0x01
  #define SPI_READ  0x03
  #define SPI_WRITE 0x02
  
  register __a1 short SpiDelay(void);
  auto u_int16 SpiSendReceiveSpi(register __a0 u_int16 datTopAligned, register __a1 s_int16 bits);
  
  #if 1
    void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
  #else
    void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value) {
        USEX(GPIO_ODATA) &= ~(SPI_xCS2|SPI_xCS);
        SpiSendReceiveSpi(SPI_WREN<<8, 8);
        USEX(GPIO_ODATA) |= SPI_xCS;
        USEX(GPIO_ODATA) &= ~SPI_xCS;
        SpiSendReceiveSpi(SPI_WRITE<<8, 8);
        SpiSendReceiveSpi(addr/*volume=26*/, 16);
        SpiSendReceiveSpi(value, 16);
        USEX(GPIO_ODATA) |= SPI_xCS|SPI_xCS2;//SPI_CLK;
    }
  #endif
  //SaveWord(26/*volume=26*/, USEX(SCI_VOL));
#endif

#ifdef SHUFFLE_PLAY
  u_int16 shuffle_offset = 0;
#endif


/*
  Called before HALT
 */
void UserHook(void) {
    register __c1 SPISTATE cachedSpiState = spiState;

    #ifdef RECORDER
      #ifdef UART_BUFFERED
        if (UartFill() >= 1) {
            /* All commands are 3 bytes long. */
            register int c = UartPeekByte();
            if (c == 'D' /*0x44 bb bb*/) {
                if (UartFill() >= 3) {
                    register int arg;
                    UartGetByte();
                    arg = UartGetByte();
                    arg = (arg<<8) | UartGetByte();
                    /* Set date */
                    created[0] = arg;
                    putch('d');
                }
            } else if (c == 'T' /*54 bb bb*/) {
                if (UartFill() >= 3) {
                    register int arg;
                    UartGetByte();
                    arg = UartGetByte();
                    arg = (arg<<8) | UartGetByte();
                    /* Set time */
                    created[1] = arg;
                    putch('t');
                }
            } else if (c == 'S' /*53 bb bb*/) {
                if (UartFill() >= 3) {
                    register int arg;
                    UartGetByte();
                    arg = UartGetByte();
                    arg = (arg<<8) | UartGetByte();
                    /* Set seconds */
                    created[2] = arg;
                    putch('s');
                }
                #ifdef VOICE_ACTIVATED
              } else if (c == 'L' /*4c bb*/) {
                  if (UartFill() >= 2) {
                      UartGetByte();
                      /*activation/deactivation level*/
                      recordThreshold = UartGetByte()*256;
                      putch('l');
                  }
              } else if (c == 'O' /*4f bb*/) {
                  if (UartFill() >= 2) {
                      UartGetByte();
                      /*timeout*/
                      recordTimeout = UartGetByte() *16;
                      putch('o');
                  }
                #endif
            } else {
                UartGetByte();
                /* unknown command */
                putch('u');
            }
        }
      #endif
    #endif/*RECORDER*/
    
    
    #if 0
      USEX(SER_DREQ) ^= 1;
      //USEX(UART_DATA) = spiState;
    #endif
    
    #ifdef SCI_UI
      /* play pause/continue */
      if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
          if (cachedSpiState != spi_pause) {
              oldSpiState = cachedSpiState;
              #ifdef UART_UI
                putch('p');//USEX(UART_DATA) = 'p';
              #endif
              cachedSpiState = spi_pause;
  
              stream_save = stream_wr_pointer;
              stream_wr_pointer = stream_rd_pointer+1;
              /* Does not matter if it is one off..
                 StreamDiff() returns the right value regardless. */
          }
      } else {
          if (cachedSpiState == spi_pause) {
          #ifdef UART_UI
                putch('c');//USEX(UART_DATA) = 'c';
          #endif
              cachedSpiState = oldSpiState;
              stream_wr_pointer = stream_save;
          }
      }
      #ifdef RECORDER
        if ((USEX(SCI_AICTRL3) & CTRL3_RECORD_ON)) {
            if (cachedSpiState != spi_waitima && cachedSpiState != spi_record) {
                record.record = 0x1234;
                goto out_of_wav;
            }
        } else {
            if (cachedSpiState == spi_waitima) {
                USEX(SCI_AICTRL0) = 0;
                goto out_of_wav;
            }
        }
      #endif/*RECORDER*/
    
    if ((USEX(SCI_AICTRL0) & 0x8000U)) {
        /* new song */
        USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
        /*NEW: clear pause and file ready bits automatically */
        USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON | CTRL3_FILE_READY | CTRL3_AT_END);
        SpiSendClocks();

        #ifdef RECORDER
      out_of_wav:
          if (cachedSpiState == spi_waitima) {
              /* wait until MMC write finished */
              while ( SpiSendReceiveMmc(-1/*0xffffU*/, 16) != 0xffffU)
                  ;
  
              SpiSendClocks();
              Disable(); //prevent modu_int from trashing the FAT
              /* Well, it would not because in VS1053 the stream buffer
                 is not used in encode mode. */
  
              mmc.fileSize = spix.size;
              #if 1 /*Fix WAV size fields */
                {
                    register u_int32 t = mmc.fileSize - 8;
                    ReadDiskSector(mmc.freeStart);
                    stream_buffer[2] = SwapWord(t);
                    stream_buffer[3] = SwapWord(t>>16);
                    t -= 0x24;
                    stream_buffer[20] = SwapWord(t);
                    stream_buffer[21] = SwapWord(t>>16);
                          
                    /* We don't care if it fails or not. */
                    MmcWrite(mmc.freeStart, stream_buffer);
                }
              #endif
              FsFatMakeFile();
              //MyReset();
          } else {
              /* especially needed for MIDI!! */
              //USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
              //cachedSpiState = spi_zeros;
              /* Discard any unread data, including CRC, before starting again */
              {
                  register int i;
                  for (i=0; i<256+2; i++)
                      SpiSendReceiveMmc(-1/*0xffff*/,16);
              }
          }
          /* Jump to new track immediately */
          #ifdef UART_UI
            putch('e');//USEX(UART_DATA) = 'e';
          #endif
          MyReset();
        #else/*RECORDER*/
        
          /* Discard any unread data, including CRC, before starting again */
          {
              register int i;
              for (i=0; i<256+2; i++)
                  SpiSendReceiveMmc(-1/*0xffff*/,16);
          }
          /* Jump to new track immediately */
          MyReset();
        #endif/*elseRECORDER*/
    }
    #endif/*SCI_UI*/
    
    #ifndef VOICE_ACTIVATED /* no keys in this mode */
      
        
          // Definitions for Standalone Legacy Keys (VS1011/VS1003/VS1053) SW1, SW2, SW3
          // These are read using special diode connections to the spi bus
      
      #define SW1 0xffffU
      #define SW2 0x8000U
      #define SW3 0x7fffU
      //#define SW4 0x4000U
      
      #ifdef ENABLE_UI
        
              // PKP: -- UI: USER INTERFACE --
          
              if (ui.Strobe) { // PKP: It's time to check the buttons and do UI calculations
                  /*
                    + SCLK = GPIDATA8
                    + XCS  = GPIDATA9
                    + SI   = GPIDATA10
                    + XDCS = GPIDATA11
                    The following code is compatible with the prototyping board buttons.
                    Code is currently 1 word longer.
                    Connecting the buttons differently 4 keys could be read directly.
                  */
                  ui.Data = ui.Strobe = 0;
                  #ifdef VOICE_ACTIVATED
                    /* No keys in voice-activation mode */
                  #else
                  
                    // PKP: -- UI: READ KEYS --
            
                    #ifdef COMPAT_KEYS // PKP: Read the legacy 3-button keys
                      if (!(USEX(GPIO_IDATA) & (1<<10))) {
                          ui.Data = SW2;
                      } else if (!(USEX(GPIO_IDATA) & (1<<11))) {
                          ui.Data = SW1;
                      } else {
                          register u_int16 t = USEX(GPIO_ODATA);
                          USEX(GPIO_ODATA) = t & ~SPI_xCS;
                          /* Should have a short wait here, but seems to work this way.*/
                          if (!(USEX(GPIO_IDATA) & (1<<11))) {
                              ui.Data = SW3;
                          } else if (!(USEX(GPIO_IDATA) & (1<<9))) {
                              ui.Data = SW3; 
                          }
                          USEX(GPIO_ODATA) = t;
                      }
                    #else // PKP: Read directly connected keys
                      //USEX(UART_DATA) = USEX(GPIO_IDATA) >> 4;
                      if (!(USEX(GPIO_IDATA) & (1<<10))) {
                          ui.Data = SW2;
                      } else if (!(USEX(GPIO_IDATA) & (1<<11))) {
                          ui.Data = SW1;
                      } else if (!(USEX(GPIO_IDATA) & (1<<9))) {
                          ui.Data = SW3;
                      } else if (!(USEX(GPIO_IDATA) & (1<<8))) {
                          ui.Data = SW1;
                      } else if ((USEX(GPIO_IDATA) & (1<<4))) {
                          ui.Data = SW2;
                      }
                    #endif
                  #endif
        
      
              // PKP: -- UI: KEY FUNCTIONS --
      
              #ifdef RECORDER/*recorder with buttons*/
                if (ui.Data == ui.Key) { //PKP: current keystate is same as previous keystate
                    if (ui.KeyTime < 12) {
                        ui.KeyTime++;
                    } else if (ui.Key) {
                        register u_int16 vol = USEX(SCI_VOL);
                        if (ui.Key == SW1) {
                            if (vol) {
                                vol -= 0x0101;
                                goto setvol;
                            }
                        } else if (ui.Key == SW2) {
                            if (vol != (s_int16)0xffffU)
                                vol += 0x0101;
                        setvol:
                            USEX(SCI_VOL) = vol;
                            #ifdef SAVE_VOLUME
                              volChanged = 1;
                            #endif/*SAVE_VOLUME*/
                        } else if (ui.Key == SW3) {
                            if (ui.KeyTime == 12) {
                                if (cachedSpiState != spi_waitima) {
                                    /* SW3 long press, goto record (if not already) */
                                    record.record = 0x1234;
                                }
                                /* If in record mode, end record mode. */
                                ui.KeyTime = -32767; /*one-shot*/
                                goto out_of_wav;
                            }
                        }
                    }
                } else {
                    if (ui.Data == 0) {
                        if (ui.KeyTime > 0 && ui.KeyTime < 12) {
                            /* was pressed < 0.75 seconds */
                            if (ui.Key == SW3) {
                                /* play pause/continue */
                                if (cachedSpiState == spi_waitima) {
                                    /* recording on, stop recording! */
                                    goto out_of_wav;
                                } else {
                                    if (cachedSpiState != spi_pause) {
                                        oldSpiState = cachedSpiState;
                                        cachedSpiState = spi_pause;
        
                                        stream_save = stream_wr_pointer;
                                        stream_wr_pointer = stream_rd_pointer+1;
                                        /* Does not matter if it is one off..
                                           StreamDiff() returns the right value
                                           regardless. */
                                    } else {
                                        cachedSpiState = oldSpiState;
                                        stream_wr_pointer = stream_save;
                                    }
                                }
                            }
                            if (ui.Key == SW1 || ui.Key == SW2) {
                                if (ui.Key == SW1) {
                                    USEX(SCI_AICTRL0) += 1; /* next file */
                                } else {
                                    /* Previous file */
                                    if (USEX(SCI_DECODE_TIME) < 5) {
                                        if (USEX(SCI_AICTRL0) == 0) {
                                            USEX(SCI_AICTRL0) = USEX(SCI_AICTRL1) - 1;
                                        } else {
                                            USEX(SCI_AICTRL0) -= 1; /* previous file */
                                        }
                                    }
                                }
                                #if defined(ENABLE_LAYER12)
                                  /* gain=0 for partial last frame (bugs) */
                                  volume[0] = volume[1] = 0;
                                #endif
                            out_of_wav:
                                if (cachedSpiState == spi_waitima) {
                                #ifdef VOICE_ACTIVATED
                                  #ifdef UART_BUFFERED
                                        putch('e');
                                  #else
                                        #ifdef USE_DEBUG
                                          puts("ending");
                                        #endif
                                  #endif
                                #endif
                                    USEX(SER_DREQ) = 0;
                                    /* wait until MMC write finished */
                                    while ( SpiSendReceiveMmc(0xffffU, 16) != 0xffffU)
                                        ;
                                    SpiSendClocks();
                                    Disable(); //prevent modu_int from trashing the FAT
                                    /* Well, it would not because in VS1053 the stream
                                       buffer is not used in encode mode. */
        
                                    mmc.fileSize = spix.size;
        
                                    #if 1 /*Fix WAV size fields */
                                      {
                                          register u_int32 t = mmc.fileSize - 8;
                                          ReadDiskSector(mmc.freeStart);
                                          stream_buffer[2] = SwapWord(t);
                                          stream_buffer[3] = SwapWord(t>>16);
                                          t -= 0x24;
                                          stream_buffer[20] = SwapWord(t);
                                          stream_buffer[21] = SwapWord(t>>16);
                                
                                          /* We don't care if it fails or not. */
                                          MmcWrite(mmc.freeStart, stream_buffer);
                                      }
                                    #endif
                                    FsFatMakeFile();
                                    #ifdef VOICE_ACTIVATED
                                      recordStarted = 0;
                                      #ifdef UART_BUFFERED
                                        putch('c'); /*created*/
                                      #else
                                        #ifdef USE_DEBUG
                                          puts("created");
                                        #endif
                                      #endif
                                      USEX(GPIO_ODATA) &= ~(1<<7);
                                    #endif
                                    MyReset();
                                }
        
                                /* especially needed for MIDI!! */
                                USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
                                cachedSpiState = spi_zeros;
                                //SpiSendClocks();
                            }
                        }
                    }
                    ui.Key = ui.Data;
                    ui.KeyTime = 0;
                }
                //ui.Data = 0;
        
        
                // PKP: Keys
        
                // SW1 Short Press: Next Song
                // SW2 Short Press: Previous Song
                // SW3 Short Press: Pause/Play
        
                // SW1 Long Press: Volume Up
                // SW2 Long Press: Volume Down
                // SW3 Long Press when Playing: Toggle Loudness
                // SW3 Long Press when Paused: Toggle Shuffle Play
        
        
              #else /* start of non-RECORDER */
                if (ui.Data == ui.Key) { // PKP: Current Keystate is same as previous Keystate
                    if (ui.KeyTime < 8) {
                        ui.KeyTime++;
                    } else // PKP: keystate has been consistent for 8 ticks - it's a long keypress
                      if (ui.Key
                      #ifdef WAIT_FOR_BUTTON
                                 && ui.KeyTime > 0
                      #endif
                        ) {
                        /* long presses */
                        register u_int16 vol = USEX(SCI_VOL);
                        if (ui.Key == SW1) {
                            if (vol) {
                                vol -= 0x0101;
                                goto setvol;
                            }
                        } else if (ui.Key == SW2) {
                            if (vol != (s_int16)0xffffU)
                                vol += 0x0101;
                        setvol:
                            USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
                            #ifdef SAVE_VOLUME
                              volChanged = 1;
                            #endif/*SAVE_VOLUME*/
                        } else {
                            if (ui.KeyTime == 8) {
                                ui.KeyTime = 9;
                                #if defined(LOUDNESS_ALWAYS_ON_OFF) && !defined(NO_RANDOM)
                                  /* if pause on -- toggle random play */
                                  USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
                                  if (cachedSpiState == spi_pause) {
                                      cachedSpiState = oldSpiState;
                                      stream_wr_pointer = stream_save;
                                  }
          
                                #else /*LOUDNESS_ALWAYS_ON_OFF*/
                                
                                  #ifdef NO_RANDOM
                                    /* toggle loudness - have LSb set when ON */
                                    USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
                                  #else/*NO_RANDOM*/
                                    if (cachedSpiState
                                    #ifdef SW3_RANDOM_PLAY
                                          !=
                                    #else
                                          ==
                                    #endif
                                        spi_pause) {
                                        /* if pause on -- toggle random play */
                                        USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
                                    } else {
                                        /* toggle loudness - have LSb set when ON */
                                        USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
                                    }
                                  #endif/*else NO_RANDOM*/
                                #endif/*else LOUDNESS_ALWAYS_ON_OFF*/
                            }
                        }
                      }
                } else { // PKP: Keystate is not the same as previous keystate
                         // PKP: Handle short keypresses
                  if (ui.Data == 0
                  #ifdef WAIT_FOR_BUTTON
                        && ui.KeyTime > 0
                  #endif
                      ) {
                      // PKP: Handle end of short keypress (current keystate ui.Data=0)
                      /* short presses */
                      if (ui.KeyTime < 8) {
                            /* was pressed < 0.5 seconds */
        
        
        
                            if (ui.Key == SW3) {
                                /* play pause/continue -- not very good for midi.. */
                                if (cachedSpiState != spi_pause) {
                                #ifndef UI_PAUSE_BEFORE_PLAY
                                      oldSpiState = cachedSpiState;
                                      cachedSpiState = spi_pause;
                                      stream_save = stream_wr_pointer;
                                      stream_wr_pointer = stream_rd_pointer+1;
                                #endif/*!UI_PAUSE_BEFORE_PLAY*/
                                } else {
                                    cachedSpiState = oldSpiState;
                                    /* restore 'stolen' data */
                                    stream_wr_pointer = stream_save;
                                }
                            } else
                            if (ui.Key == SW1 || ui.Key == SW2) {
                                register u_int16 song = USEX(SCI_AICTRL0);
                                if (ui.Key == SW1) {
                                    song++; /* next file */
                                } else
                                #ifndef UI_PAUSE_BEFORE_PLAY
                                      if (USEX(SCI_DECODE_TIME) < 5)
                                #endif
                                    {
                                    if (song == 0) {
                                        song = USEX(SCI_AICTRL1)-1;
                                    } else { /* previous file */
                                        song--;
                                    }
                                    /* else the same file */
                                }
                                USEX(SCI_AICTRL0) = song;
        
                                /* especially needed for MIDI!! */
                                #if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
                                  USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
                                #endif
                                cachedSpiState = spi_zeros;
                                #if defined(ENABLE_LAYER12)
                                  /* gain=0 for partial last frame (bugs) */
                                  volume[0] = volume[1] = 0;
                                #endif
                            }
                      }//(ui.keyTime was < 8)
                  }
                    ui.Key = ui.Data;
                    ui.KeyTime = 0;
                }
        
                #if defined(SAVE_VOLUME)
                  /*Save volume when volume button is released. */
                  if (ui.Data == 0 &&
                      volChanged &&
                      cachedSpiState == spi_seek) {
                      volChanged = 0;
                      SaveWord(26/*volume=26*/, USEX(SCI_VOL));
                  }
                #endif/*SAVE_VOLUME*/
              
              //ui.Data = 0;
        #endif /* !RECORDER */
        }
      #endif /*USE_UI*/
    #else
    
    #endif /*!VOICE_ACTIVATED*/
    
    
    #ifdef PLAYTIME
      if (USEX(SCI_DECODE_TIME) > PLAYTIME && cachedSpiState != spi_zeros) {
          USEX(SCI_AICTRL0)++;
          #if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
            USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
          #endif
          cachedSpiState = spi_zeros;
      }
    #endif/*PLAYTIME*/
    
    
    #ifdef RECORDER /* 88 words */
      if (cachedSpiState == spi_waitima) {
          if (SpiSendReceiveMmc(-1/*0xffffU*/, 16) == 0xffffU) {
              /* MMC Ready. */
              SpiSendClocks();
  
              /* Note: byteswap required for linear 16-bit data */
              if (record.fill >= 256) { /* Data in second FIFO */
                  WritePhysicalSectorYSwap(spix.sector<<
                  #ifdef USE_HC
                                             hcShift,
                  #else
                                             9,
                  #endif
                                           mallocAreaY+record.rd);
  
                  spix.sector++;
                  spix.size += 512; /* in bytes! */
                  record.fill -= 256;
                  record.rd += 256;
                  if (record.rd >= sizeof(mallocAreaY))
                      record.rd -= sizeof(mallocAreaY);
              }
              if (spix.size >= 0x7ffffe00 || /*max size about 1 day at 24kHz*/
                  spix.size/512 >= mmc.freeSize
                  #ifdef VOICE_ACTIVATED
                    || (recordStarted && recordCount == 0)
                  #endif
                  ) {
                  /* disk full, end recording. */
                  #if defined(VOICE_ACTIVATED) && !defined(SCI_UI)
                out_of_wav:
                    if (cachedSpiState == spi_waitima) {
                    #ifdef VOICE_ACTIVATED
                      #ifdef UART_BUFFERED
                            putch('e');
                      #else
                            #ifdef USE_DEBUG
                              puts("ending");
                            #endif
                      #endif
                    #endif
                        USEX(SER_DREQ) = 0;
                        /* wait until MMC write finished */
                        while ( SpiSendReceiveMmc(0xffffU, 16) != 0xffffU)
                            ;
                        SpiSendClocks();
                        Disable(); //prevent modu_int from trashing the FAT
                        /* Well, it would not because in VS1053 the stream
                           buffer is not used in encode mode. */
                        
                        mmc.fileSize = spix.size;
                        
                        #if 1 /*Fix WAV size fields */
                          {
                              register u_int32 t = mmc.fileSize - 8;
                              ReadDiskSector(mmc.freeStart);
                              stream_buffer[2] = SwapWord(t);
                              stream_buffer[3] = SwapWord(t>>16);
                              t -= 0x24;
                              stream_buffer[20] = SwapWord(t);
                              stream_buffer[21] = SwapWord(t>>16);
                              
                                      /* We don't care if it fails or not. */
                              MmcWrite(mmc.freeStart, stream_buffer);
                          }
                        #endif
                        FsFatMakeFile();
                        #ifdef VOICE_ACTIVATED
                          recordStarted = 0;
                          #ifdef UART_BUFFERED
                            putch('c'); /*created*/
                          #else
                            #ifdef USE_DEBUG
                              puts("created");
                            #endif
                          #endif
                          USEX(GPIO_ODATA) &= ~(1<<7);
                        #endif
                        MyReset();
                    }
                  #endif
                  goto out_of_wav;
              }
          }
  
          /* Fill second FIFO */
  #define READ_SIZE 32 /*words*/
          {
              register s_int16 imaFil = (bssWp - bssRp) & (IMA_ENCODE_BSSPACE-1);
              if (imaFil >= READ_SIZE) {
                  s_int16 ne = record.wr + READ_SIZE;
                  if (ne >= sizeof(mallocAreaY))
                      ne -= sizeof(mallocAreaY);
                      #ifndef VOICE_ACTIVATED /* Well, not needed for non-voice-activated either */
                    if (ne != record.rd)
                      #endif
                  {
                  #ifdef VOICE_ACTIVATED
                        memcpyYY(mallocAreaY+record.wr, bssRp, READ_SIZE);
                        bssRp += READ_SIZE;
                        if (bssRp >= g_dcthi+IMA_ENCODE_BSSPACE) {
                            bssRp -= IMA_ENCODE_BSSPACE;
                        }
                        if ((USEX(GPIO_IDATA) & (1<<6))) {
                            /* Disable start of recording or end recording */
                            recordCount = 0;
                        } else {
                            register __y s_int16 *p = mallocAreaY+record.wr;
                            register int i;
                            register u_int16 th = 0;
                            for (i=0;i<READ_SIZE;i++) {
                                th |= abs(*p);
                                p++;
                            }
                            /* thresholds < 256 -> 0 */
                            th = (u_int32)th * th >> 16;
                            if (th < recordThreshold) {
                                /* Below the threshold */
                            } else {
                                /* Above threshold, start saving */
                                if (recordStarted) {
                                    recordCount = recordTimeout;
                                } else {
                                    if (++recordCount > 2) {
                                        USEX(GPIO_ODATA) |= (1<<7);
                                        recordStarted = 1;
                                        #ifdef UART_BUFFERED
                                          putch('r');
                                        #else
                                          #ifdef USE_DEBUG
                                            USEX(INT_ENABLE) &= ~(1<<INT_EN_RX);
                                            puts("record");
                                          #endif
                                        #endif
                                    }
                                }
                            }
                            if (recordStarted) {
                                record.wr = ne;
                                record.fill += READ_SIZE;
                            }
                        }
                  #else
                        memcpyYY(mallocAreaY+record.wr, bssRp, READ_SIZE);
                        bssRp += READ_SIZE;
                        if (bssRp >= g_dcthi+IMA_ENCODE_BSSPACE) {
                            bssRp -= IMA_ENCODE_BSSPACE;
                        }
                        record.wr = ne;
                        record.fill += READ_SIZE;
                  #endif
                  }
                  #ifndef SCI_UI
                    ui.Cnt += READ_SIZE; /* Mono linear 16-bit */
                    if (ui.Cnt >= RECORD_FS/16) {
                        ui.Cnt -= RECORD_FS/16;
                        #ifdef VOICE_ACTIVATED
                          if (recordStarted && recordCount) {
                              recordCount--;
                          }
                        #endif
                        ui.Timer++;
                        ui.Strobe = 1;
                        /* for timebase, but may be too fast or slow.. */
                    }
                  #endif
              }
          }
  
          #if defined(LOWER_LOOPBACK_VOL) /* lower loopback volume */
            volume[0] = LOWER_LOOPBACK_VOL;
            volume[1] = LOWER_LOOPBACK_VOL;
          #endif
      } else if (cachedSpiState == spi_record) {
  #define BYTESWAP(a) (((u_int16)(a)>>8)|((a)<<8))
      #if 0
            static __y u_int16 waveHeader[] = {
                /* 8000Hz 1-ch 16-bit signed WAV */
                0x5249,0x4646,0x2400,0x0000,0x5741,0x5645,0x666d,0x7420,
                0x1000,0x0000,0x0100,0x0100,
                BYTESWAP(RECORD_FS), /* 0x401f*/
                0x0000,
                BYTESWAP(RECORD_FS*2), /*0x803e*/
                0x0000,
                0x0200,0x1000,0x6461,0x7461,0xffff,0xffff,0x0000,0x0000,
                0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
                /* 10 zero-samples inserted to align to 32-word boundary */
            };
            /* Currently Mono only */
            waveHeader[12] = SwapWord(USEX(SCI_AICTRL0)); /* samplerate */
            waveHeader[14] = SwapWord(USEX(SCI_AICTRL0)*2); /* byterate */
            waveHeader[15] = SwapWord((u_int32)USEX(SCI_AICTRL0)*2 >> 16);
    
            memset(stream_buffer, 0, 512/2);
            memcpyYX(stream_buffer, waveHeader, sizeof(waveHeader));
            MmcWrite(mmc.freeStart, stream_buffer);
    
            spix.size = 512; /* calculate file size, in bytes! */
            /* Skip WAV header, start from second sector. */
            spix.sector = mmc.freeStart + 1;
            cachedSpiState = spi_waitima;
      #else
            /* Data save performs byteswap, so header is byte-swapped! */
            static __y u_int16 waveHeader[] = {
                /* 8000Hz 1-ch 16-bit signed WAV */
                0x4952,0x4646,0x0024,0x0000,0x4157,0x4556,0x6d66,0x2074,
                0x0010,0x0000,0x0001,0x0001,
                RECORD_FS, /* 0x401f*/
                0x0000,
                RECORD_FS*2, /*0x803e*/
                ((RECORD_FS*2)>>16), /*0x0000*/
                0x0002,0x0010,0x6164,0x6174,0xffff,0xffff,0x0000,0x0000,
                #ifndef VOICE_ACTIVATED
                  0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
                #endif
                /* 10 zero-samples inserted to align to 32-word boundary */
            };
            /* Currently Mono only */
            #ifndef VOICE_ACTIVATED /*fields already correct because AICTRL0==RECORD_FS*/
              waveHeader[12] = USEX(SCI_AICTRL0); /* samplerate */
              waveHeader[14] = USEX(SCI_AICTRL0)*2; /* byterate */
              waveHeader[15] = (u_int32)USEX(SCI_AICTRL0)*2 >> 16;
            #endif
            memsetY(mallocAreaY, 0, 512/2);
            memcpyYY(mallocAreaY, waveHeader, sizeof(waveHeader));
            record.fill = record.wr = READ_SIZE;
            spix.size = 0; /* calculate file size, in bytes! */
            /* Skip WAV header, start from second sector. */
            spix.sector = mmc.freeStart;
            cachedSpiState = spi_waitima;
      #endif
      }
    #endif/*RECORDER*/
    
    if (cachedSpiState == spi_file) {
        /* extra byte read if odd file size */
        spix.size = (fatInfo.fileSize+1)/2; /* in words! */
        cachedSpiState = spi_seek;
setfrag:
        spix.fragment = &fragments[0];
        spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
        spix.fragSize = fragments[0].size; /* *fatInfo.sectorsPerCluster*/
    }
    #if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
      if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG &&
          USEX(SCI_HDAT1) == 0x7665/*(('v'<<8)|'e')*/) {
          if (USEX(SCI_DECODE_TIME) < lastDecodeTime)
              USEX(SCI_DECODE_TIME) = lastDecodeTime;
      }
    #endif /*SCI_UI and FIX_WAV_DECODETIME*/
    
    if (cachedSpiState == spi_seek) {

    #ifndef SCI_UI
          /* In case we are just trying to find something to play,
             also read keys once in a while.. */
          ui.Cnt += 64;
          if (ui.Cnt > hwSampleRate/16) {
              ui.Cnt -= hwSampleRate/16;
              ui.Timer++;
              ui.Strobe = 1;
          }
    #endif
    
    #ifdef USE_HC
          MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
                     (spix.sector & 0x7fffffffUL) << hcShift);
    #else
          MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(spix.sector<<9));
    #endif
        cachedSpiState = spi_waitdata;
        spiCnt = 0;
        NextSector();
        #ifdef HDATCHECK
          if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
              /* if file not detected at HDATCHECK/2 kB, skip the rest */
              goto end_of_file;
          }
        #endif
    }

    #ifndef VOICE_ACTIVATED
      /*TODO: handle spi_waitdata and spi_data
              in timer interrupt to increase throughput for FLAC */
      if (cachedSpiState == spi_waitdata) {
          if (SpiSendReceiveMmc(0xff00,8) == 0xfe) {
              cachedSpiState = spi_data;
              spiCnt = 0;
              #ifdef TIMEOUT
                timeOut = 0;
              #endif
          } else {
              if ((s_int16)++spiCnt < 0)
                  goto goreset; //MyReset(); /* timeout */
          }
      } else if (cachedSpiState == spi_data) {
          while (spix.size > 0) {
              register int i, j = 32;
              register u_int16 *wr;
              #if !defined(RECORDER) && defined(USE_FLAC)
                if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
                    register s_int16 t = stream_wr_pointer - stream_rd_pointer;
                    if (t < 0)
                        t += 0x1800;
                    if (t > 0x1800/*STREAM_BUFFER_SZ*/-40)
                        break;
                    wr = (void *)stream_wr_pointer;
                    if (j > spix.size)
                        j = spix.size;
                    for (i=0;i<j;i++) {
                        *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
                        if (wr >= &stream_buffer[0x1800/*STREAM_BUFFER_SZ*/])
                            wr = stream_buffer;
                    }
                } else
              #endif
              {
                  if (StreamDiff() >= STREAM_BUFFER_SZ-40)
                      break;
                  wr = (void *)stream_wr_pointer;
                  if (j > spix.size)
                      j = spix.size;
                  for (i=0;i<j;i++) {
                      *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
                      if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
                          wr = stream_buffer;
                  }
              }
              stream_wr_pointer = (void *)wr;
              spix.size -= j;//32;
              spiCnt += 32;
              if (spiCnt >= 256) {
                  /* 256 words read, get next block in file */
                  cachedSpiState = spi_seek;
                  SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
                  break;
              }
              #ifdef TIMEOUT
                timeOut = 0;
              #endif
              /* watchdog clearing here doesn't work with midi */
          }
  
          #ifdef TIMEOUT
            if ((s_int16)++timeOut == -1) {
                spix.size = 0;
            }
          #endif
    
    #ifdef ENABLE_WATCHDOG
          USEX(WDOG_RESET) = WDOG_RESET_VAL;
    #endif
        if (spix.size <= 0) {
end_of_file:
            SpiSendClocks();

            #ifdef SCI_UI
              if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG){
                  /* restart the same file immediately */
                  #if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
                    lastDecodeTime = USEX(SCI_DECODE_TIME)+1;
                  #endif /*SCI_UI and FIX_WAV_DECODETIME*/
                  cachedSpiState = spi_file;
                  #ifdef UART_UI
                    putch('l');//USEX(UART_DATA) = 'l';
                  #endif
              } else
            #endif/*SCI_UI*/
            {
            #if 0
                  USEX(SCI_AICTRL0)++; /* next file */
                  /* Just zeros, no outofwav */
                  cachedSpiState = spi_zeros;
            #else
                  #ifdef SCI_UI
                    /*NEW: pause after play */
                    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
                        CTRL3_PAUSE_AFTER_PLAY) {
                        USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
                    }
                  #else
                    {
                        register int m = (USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK);
                        if (m == CTRL3_LOOP_SONG) {
                            /* restart the same file immediately */
                            cachedSpiState = spi_file;
                        } else if (m == CTRL3_PAUSE_AFTER_PLAY) {
                            oldSpiState = spi_zeros;
                            cachedSpiState = spi_pause;
                        } else {
                            USEX(SCI_AICTRL0) += 1; /* next file */
                            cachedSpiState = spi_zeros;
                            #if 1
                              if (USEX(SCI_HDAT1) == (('f'<<8) | 'L'))
                                  USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
                            #endif
                        }
                    }
                  #endif
            #endif
            }
        }
        #ifdef NO_WMA /* reject WMA files */
          if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
              USEX(SCI_AICTRL0) += 1; /* next file */
              goto goreset;//MyReset();
          }
        #endif
        #if defined(NO_AAC) && !defined(SCI_UI) && !defined(RECORDER)
          {
              register u_int16 t = USEX(SCI_HDAT1);
              //0x4154(AT) 0x4144(AD) 0x4d34(M4)
              if (t == 0x4154 || t == 0x4144 || t == 0x4d34) {
                  USEX(SCI_AICTRL0) += 1; /* next file */
                  goto goreset;//MyReset();
              }
          }
        #endif
    } else if (cachedSpiState == spi_zeros) {
        /* Workaround for WMA resync after stream ended! */
        parametric_x.resync = 0;

        if (StreamDiff() < STREAM_BUFFER_SZ-40) {
            register int i, z = parametric_x.endFillByte;
            register u_int16 *wr = (void *)stream_wr_pointer;
            for (i=0;i<32;i++) {
                *wr = z;
                if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
                    wr = stream_buffer;
                } else {
                    wr++;
                }
                ++spiCnt;
                stream_wr_pointer = (void *)wr;
            }
            /* watchdog clearing here doesn't work with midi */
            #ifdef TIMEOUT
              timeOut = 0;
            #endif
        }
        #ifdef TIMEOUT
          if ((s_int16)++timeOut < 0) {
  //putch(USEX(SCI_HDAT1)>>8);
  //putch(USEX(SCI_HDAT1));
              goto goreset;
          }
        #endif
        
        if (USEX(SCI_HDAT1) == 0 ||
        #if 1
              USEX(SCI_HDAT1) == 0x574d ||
        #endif
            spiCnt >= 256+2048) { /* was left to max 256 by spi_data */
            /* 2048 words instead of bytes */
goreset:
            /* Discard any unread data, including CRC, before starting again */
            {
                register int i;
                for (i=0; i<256+2; i++)
                    SpiSendReceiveMmc(-1/*0xffff*/,16);
            }
            SpiSendClocks();

            #ifdef UART_UI
              putch('e');//USEX(UART_DATA) = 'e';
            #endif
            MyReset();
        }
        #ifdef ENABLE_WATCHDOG
          goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
        #endif
    } else {
    #ifdef ENABLE_WATCHDOG
  wdogack:
          USEX(WDOG_RESET) = WDOG_RESET_VAL;
    #endif
    }
    #endif /* VOICE_ACTIVATED*/
    
    #ifdef SCI_UI
      USEX(SER_DREQ) = 0;
    #else
    
      #ifndef NO_DREQ_LED
        /* 15 words */
        {
            register int t = 0;
    
            if ((ui.Timer & 0x18) == 0) {
                /* DREQ indicates random play status */
                #ifdef NO_RANDOM
                  t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
                #else
                  t = USEX(SCI_AICTRL3);  /* lowest bit is global random play */
                  #ifdef LOUDNESS_ALWAYS_ON_OFF
                    /* indicate random play */
                  #else /*LOUDNESS_ALWAYS_ON_OFF*/
                  
                    if (cachedSpiState
                    #ifdef SW3_RANDOM_PLAY
                          ==
                    #else
                          !=
                    #endif
                        spi_pause)
                        t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
                  #endif /*else LOUDNESS_ALWAYS_ON_OFF*/
                #endif /*else NO_RANDOM*/
            }
            USEX(SER_DREQ) = t;
        }
      #endif/*!NO_DREQ_LED*/
    #endif/*!SCI_UI*/
    
    spiState = cachedSpiState;

}


#ifdef VS1053
  void NewSinTest(void);
  auto int PatchFrame(void);
#endif


auto void MyMain(register s_int16 __a1 fullReset) {
    #ifdef VOICE_ACTIVATED
      putch('1');
    #endif

    #ifdef UART_BUFFERED
      UartFill();
    #endif

    #ifdef VOICE_ACTIVATED
      putch('2');
    #endif

    #ifdef VS1053
      /* Switch */
      USEX(VS1053_IROM4) &= ~(IROM4_ENABLE);
    #endif

    #if defined(SCI_UI) && defined(ENABLE_WATCHDOG)
      USEX(WDOG_CONFIG) = 1000;
      USEX(WDOG_RESET) = WDOG_RESET_VAL;
    #endif

    #if defined(SCI_UI)
      USEX(SCI_AICTRL1) = 0;
    #endif

    USEX(SCI_MODE) = SciModeOr() | (USEX(SCI_MODE) & ~(1<<SCIMB_OUT_OF_WAV));
    USEX(GPIO_ODATA) = SPI_xCS|SPI_CLK|SPI_xCS2|(1<<6); /*use same= -1 word*/
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2|(1<<6)
        #ifdef VOICE_ACTIVATED
          | (1<<7)
        #endif        
        #if defined(VS1053) && defined(ENABLE_I2S)
          /* Enable I2S */
          | 0xf0;
      USEX(I2S_CONTROL) = 0x0c
        #endif
        ;

    #ifdef VOICE_ACTIVATED
      putch('3');
    #endif
        
    InitHardware(); /* no longer clears memory in vs1053 */
    #ifdef RECORDER
      record.record <<= 1; /* 0x1234 to 0x2468 */
    #endif
    
    #ifndef USE_DEBUG
      /* Allow reflashing without MMC/SD inserted. */
      USEX(INT_ENABLE) = (1<<INT_EN_RX);
    #endif
    InitLayer3();
    InitCommon();
    //InitReaders();
    InitDecode();
    //InitGeneral();
    //AudioInit();

    #ifdef VOICE_ACTIVATED
      putch('4');
    #endif
    
    #if !defined(ENABLE_LAYER12)
      /* If not powerup reset, re-enable analog immediately. */
      if (!fullReset) {
          USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
      }
      #if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
        else {
            /*analog power off and analog driver powerdown */
            USEX(SCI_STATUS) |= (1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2);
        }
      #endif
    #endif
    
    
    #ifdef RECORDER

      #ifdef ENABLE_UI
        applAddr = ApplAddrRec; //Must call idle handler
      #endif/*ENABLE_UI*/
      #ifdef SCI_UI
        applAddr = ApplAddrSciRec; //Must call idle handler
      #endif/*ENABLE_UI*/
    
    #else/*RECORDER*/

      #if defined(SCI_UI) && defined(VS1053) && defined(ENABLE_HIGHREF)
        applAddr = ApplAddrRef;
      #endif
      #ifdef ENABLE_UI
        #if defined(VS1053) && defined(ENABLE_HIGHREF)
          applAddr = ApplAddrButRef;
        #else
          applAddr = ApplAddr;
        #endif
      #endif/*ENABLE_UI*/
    
    #endif/*elseRECORDER*/

    
    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
    //USEX(INT_ENABLE) = 0;
    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
    //USEX(INT_GLOB_ENA) = 0;

    InitReadFrame();

    #ifdef VOICE_ACTIVATED
      putch('5');
    #endif


    /* wake up MMC! -- If this takes too long, watchdog will bite. */
    /* inlining InitStorage() saves 18 words */
    /*
      Kingston 128M (Korea) -- supports interrupted read
      Cmd(7B00000000)
      00FF               (timeout waiting for reply)
      Cmd(4000000000)
      000D               (crc error, invalid command)
      not valid idle
      Cmd(7B00000000)
      0005               (crc error)
      Cmd(4000000000)
      0001               (ok)
    */

    #ifdef SAVE_POSITION
      if (USEX(SCI_AICTRL0)) /* 9 words */
          SaveWord(28/*AICTRL0*/, 0); //3 words
    #endif
    
    #ifdef VOICE_ACTIVATED
      putch('6');
    #endif

    
    //while (1)
    {
        register s_int16 i, cmd;

	// Start the MMC/SD card
    tryagain: 

	// MMC reset
	USEX(GPIO_DDR) |= (1 << 6);
	USEX(SER_DREQ) = 0;
	{
	  u_int16 gpio0_save = USEX(GPIO_ODATA);
	  USEX(GPIO_ODATA) = 1; //all 0 except eecs
	  for (i=0; i<32000; i++) {
	    USEX(GPIO_ODATA) = 1; //all 0 except eecs
	  }
	  USEX(GPIO_ODATA) = gpio0_save | (1 << 6);
	}
      

        #ifdef VOICE_ACTIVATED
          putch('i');
        #endif
	  for (i=0; i<512; i++) {
	    SpiSendClocks();
	  }


        /* MMC Init, command 0x40 should return 0x01 if all is ok. */
        i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
        if (i != 1)
            goto tryagain;//continue; /* No valid idle response */

        cmd = MMC_SEND_OP_COND|0x40;
        #ifdef USE_HC
          //MmcCommand(MMC_CRC_ON_OFF/*CMD59*/|0x40, 0);
          if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
              == 1) {
              /* MMC answers: 0x05 SD answers: 0x01 */
              /* Should we read the whole R7 response? */
              //SpiSendReceiveMmc(-1, 32);
              cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
          }
        #endif /*USE_HC*/

        /* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
        //i = 0; /* i is 1 when entered .. but does not make the code shorter*/
        while (1) {
            register int c;
            #ifdef USE_HC
              if (cmd == (0x40|41)) {
                  MmcCommand(0x40|55/*CMD55*/,0);
                  c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0x40000000UL);
              } else
            #endif
            c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0);
            if (c == 0)
                break;
            if (++i >= 12000/*< 0*/ || c != 1) {
            #ifdef USE_DEBUG
                  puts("again");
            #endif
                goto tryagain; /* Not able to power up mmc */
            }
        }

        #ifdef USE_HC
          hcShift = 9;
          i = 0x0246;
          /* PRETEC's 1G MMC does not seemd to handle CMD58 here! */
          /* Support HC for SD, but not for MMC! */
          if (cmd == (0x40|41) &&
              MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
              if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
                  /* OCR[30]:CCS - card capacity select */
                  /* HighCapacity! */
                  /* ldc 9,D0 (0x0246) <-> ldc 0,D0 (0x0006) */
                  i = 6;
                  hcShift = 0;
              }
              //SpiSendReceiveMmc(-1, 16);
          }
          WriteToProgramRam((u_int16)ReadDiskSector+2, 0, i);
        #endif /*USE_HC*/
        
        #if 0
          /* Set Block Size of 512 bytes -- default for at least HC */
          /* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
          /* Must check return value! (MicroSD restart!) */
          if (MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0)
              goto tryagain;
          /* All OK return */
        #endif
    }

    // Init File System (FAT)

    /* Detects cards without partition table */
    if (InitFileSystem() == 0) {

    #ifdef RECORDER
      #ifndef SCI_UI
            ui.Key = 0;
            ui.KeyTime = -32767; /*Ignore record button release*/
      #endif
    
        if (
        #ifdef VOICE_ACTIVATED
              1 ||
        #endif
            record.record == 0x2468) {
            /* If record mode, re-enable analog power+drivers immediately. */
            USEX(SCI_STATUS) &= ~((1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2));
            memset(&record, 0, sizeof(record));
            //record.record = 0;
            //record.rd = record.wr = 0;

	    USEX(SER_DREQ) = 0;
            FatPrepareForSaving();
            if (mmc.dirLine >= 0) {
                spiState = spi_record;
                #ifdef UART_UI
                  putch('r');//USEX(UART_DATA) = 'r';
                #endif
                #ifdef VOICE_ACTIVATED
                  putch('w');
                #endif
                
                /* Lower monitor volume during recording,
                   default values will be restored by InitHardware(). */
                /* does not work because StereoCopy not used. */
                //volume[0] = volume[1] = -128;
                /* This one works! */
                USEX(DAC_VOL) = 0x0808;

                USEX(SCI_AICTRL0) = RECORD_FS;
                USEX(SCI_AICTRL1) = USEX(SCI_WRAM);     /* auto gain */
                USEX(SCI_AICTRL2) = USEX(SCI_WRAMADDR); /* max gain */
                USEX(SCI_AICTRL3) = 4|2; //mode: linear, mono left channel
                USEX(SCI_MODE) |= RecordMode();
                USEX(SCI_MODE) &= ~(1<<SCIMB_LINE);

                ImaEncode(); /* loop there until reset */
            }
        }
    #endif/*RECORDER*/
        
        
    #ifndef VOICE_ACTIVATED
      #ifdef SCI_UI
	if ((USEX(SCI_AICTRL3) & CTRL3_NO_NUMFILES)) {
	  USEX(SCI_AICTRL1) = 0x7fff;
	} else
      #endif/*SCI_UI*/
	  {
	    /* determine number of playable files */
	    USEX(SCI_AICTRL1) = OpenFile(0xffffU);
	    //putch(USEX(SCI_AICTRL1)>>8);
	    //putch(USEX(SCI_AICTRL1));
          }
	
      #if !defined(NO_RANDOM) /* if random play feature active */
	if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
            #ifdef SHUFFLE_PLAY /*shuffle 23 words*/
	  USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
            #else
                  /* random play (not shuffle!) */
	  USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL1));
            #endif
            }
          #endif
        
        if (OpenFile(USEX(SCI_AICTRL0)
        #ifdef SCI_UI
                        &= 0x7fffU
                       //max 32768 files in SCI-controlled version
        #endif
            ) < 0) {
            #if defined(SKIP_ID3V2) /* 48 words */
              /* check for ID3v2 */
              /* An ID3v2 tag can be detected with the following pattern:
                 $49 44 33 yy yy xx zz zz zz zz */
              ReadDiskSector(fragments[0].start);
              if ((GetLong(0) & 0xffffffL) == 0x334449) {
                  /* skip in 16kB resolution, assume tags are smaller than 2MB */
                  /* 0.5kB resolution now */
                  register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
                  /* guard against small fragments */
                  if (skip >= fragments[0].size)
                      skip = fragments[0].size-1;
                  fragments[0].start += skip;
                  fragments[0].size  -= skip;
                  fatInfo.fileSize -= (u_int32)skip * 512;
              }
            #endif
            
            #if defined(SCI_UI)
              if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
                  CTRL3_PAUSE_BEFORE_PLAY) {
                  /* Mark file ready and go to pause state in UserHook */
                  USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
              }
              /*NEW: set FILE_READY always */
              USEX(SCI_AICTRL3) = CTRL3_FILE_READY | 
                  (USEX(SCI_AICTRL3) & ~CTRL3_AT_END);
            #endif
            
            #define PAUSE_AT_POWERON
            #if defined(PAUSE_AT_POWERON) /*pause at power-on, then use pause after play*/
              if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
                  CTRL3_PAUSE_BEFORE_PLAY) {
                  USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY;
                  oldSpiState = spi_file;
                  spiState = spi_pause;
                  stream_save = stream_wr_pointer;
              } else {
                  spiState = spi_file;
              }
            #else /*PAUSE_AT_POWERON*/
            
              #if defined(UI_PAUSE_BEFORE_PLAY) && defined(ENABLE_UI)
                oldSpiState = spi_file;
                spiState = spi_pause;
                stream_save = stream_wr_pointer;
              #else
                spiState = spi_file;
              #endif
            #endif /*else PAUSE_AT_POWERON*/
            
            
            #ifdef HDATCHECK
              hdatCheck = 0;
            #endif
        } else {
        gofirst:
            USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
            MyReset();
        }
    } else {
        /* Initialize again */
        goto gofirst;//MyReset();
    #else /*!VOICE_ACTIVATED*/
    } else {
        /* Initialize again */
        MyReset();
    #endif /*else!VOICE_ACTIVATED*/
    }

    #ifndef VOICE_ACTIVATED
      
        #if !defined(RECORDER) && defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
          /*otherwise keep it in powerdown until MMC found*/
          USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
        #endif
      
    
    #ifdef SAVE_POSITION
      SaveWord(28/*AICTRL0*/, USEX(SCI_AICTRL0));
    #endif
    
    
    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
    USEX(SCI_DECODE_TIME) = 0;

    #if 0 && defined(VS1053) //Sine Test
      USEX(SCI_VOL) = 0;
      SetRate(44100U);
      USEX(SCI_AICTRL0) = 2972; /*1kHz*/
      USEX(SCI_AICTRL1) = 2972;
      CallIROM4(NewSinTest);
    #endif
    
    parametric_x.resync = 32767;

    #if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
      lastDecodeTime = 0;
    #endif /*SCI_UI and FIX_WAV_DECODETIME*/
    
    #if 1
      stream_rd_pointer = stream_wr_pointer = stream_buffer;
      stream_rd_odd = 0;
    #endif
    while (1) {
        if (
        #if !defined(RECORDER) && defined(USE_FLAC)
              !PatchFrame()
        #else
              !ReadFrame()
        #endif
            ) {
            play_frame();
        } else {
            InitReadFrame();
        }
    }
    #endif /*!VOICE_ACTIVATED*/
}
